from stitch_core import compress, rewrite, StitchException, from_dreamcoder, Abstraction
import json

# simple test
programs = ["(a a a)", "(b b b)"]
res = compress(programs, iterations=1)
assert res.rewritten == ['(fn_0 a)', '(fn_0 b)']
assert res.abstractions[0].body == '(#0 #0 #0)'

# rewriting test
programs_to_rewrite = ["(c c c)", "(d d d)"]
assert rewrite(programs_to_rewrite, res.abstractions) == ['(fn_0 c)', '(fn_0 d)']

# example from Overview section of the Stitch paper (https://arxiv.org/abs/2211.16605)
programs = [
    "(lam (+ 3 (* (+ 2 4) 2)))",
    "(lam (map (lam (+ 3 (* 4 (+ 3 $0)))) $0))",
    "(lam (* 2 (+ 3 (* $0 (+ 2 1)))))"
]
res = compress(programs, iterations=1, max_arity=2)
assert res.abstractions[0].body == '(+ 3 (* #1 #0))'
assert res.rewritten == [
    '(lam (fn_0 2 (+ 2 4)))',
    '(lam (map (lam (fn_0 (+ 3 $0) 4)) $0))',
    '(lam (* 2 (fn_0 (+ 2 1) $0)))'
    ]

# loading from a file
with open('data/cogsci/nuts-bolts.json','r') as f:
    programs = json.load(f)
res = compress(programs, iterations=3, max_arity=3)
assert res.abstractions[0].body == '(T (repeat (T l (M 1 0 -0.5 (/ 0.5 (tan (/ pi #1))))) #1 (M 1 (/ (* 2 pi) #1) 0 0)) (M #0 0 0 0))'
assert res.abstractions[1].body == '(repeat (T (T #2 (M 0.5 0 0 0)) (M 1 0 (* #1 (cos (/ pi 4))) (* #1 (sin (/ pi 4))))) #0 (M 1 (/ (* 2 pi) #0) 0 0))'
assert res.abstractions[2].body == '(T (T c (M 2 0 0 0)) (M #0 0 0 0))'

# dreamcoder format
with open('data/dc/origami/iteration_0_3.json','r') as f:
    dreamcoder_json = json.load(f)
kwargs = from_dreamcoder(dreamcoder_json)
res = compress(**kwargs, iterations=3, max_arity=3)
assert res.abstractions[0].body == '(if (empty? (cdr #0)) #2 (#1 (cdr #0)))'


# StitchException: passing in an argument that doesn't actually exist
try:
    out_json = compress(programs, iterations=3, nonexistant_arg=213879)
    assert False, "Should have thrown an exception"
except StitchException as e:
    pass

# StitchException: malformed programs (or any other panic in the rust backend)
bad_programs = ["(a a a"]
try:
    out_json = compress(bad_programs, iterations=3)
    assert False, "Should have thrown an exception"
except StitchException as e:
    # print(e)
    pass

# TypeError: passing in an argument of the wrong type, so that conversion to strongly typed Rust fails
bad_programs = 4
try:
    out_json = compress(bad_programs, iterations=3, loud_panic=True)
    assert False, "Should have thrown an exception"
except TypeError as e:
    # print(e)
    pass



# import the contents of the Rust library into the Python extension
from .stitch_core import compress_backend,rewrite_backend
from typing import Optional, Dict, List, Any
import json


class StitchException(Exception):
    pass



class Abstraction:
    def __init__(self, name: str, body: str, arity: int):
        self.name = name
        self.body = body
        self.arity = arity

class CompressionResult:
    def __init__(self, json: Dict[str,Any]):
        self.abstractions = [Abstraction(body=abs["body"], name=abs["name"], arity=abs["arity"]) for abs in json["abstractions"]]
        self.rewritten = json['rewritten']
        self.json = json


def from_dreamcoder(json: Dict[str,Any]):
    """
    Takes a dreamcoder-style dictionary and returns a dictionary of arguments to pass to stitch.compress()
    """

    frontiers = json["frontiers"]
    anonymous_abstractions = [production["expression"] for production in json["DSL"]["productions"] if production["expression"].startswith("#")]
    anonymous_abstractions.sort(key=len)
    anonymous_to_named = [(f"dreamcoder_abstraction_{i}", anonymous) for (i,anonymous) in enumerate(anonymous_abstractions)]

    programs = []
    tasks = []
    for i, frontier in enumerate(frontiers):
        task = frontier.get("task", str(i))
        for program in frontier["programs"]:
            program = program["program"]
            # replace #(lambda ...) with fn_2 etc. Start with highest numbered fn to avoid mangling bodies of other fns.
            for (name, anonymous) in reversed(anonymous_to_named):
                program = program.replace(anonymous, name)
            assert '#' not in program
            programs.append(program)
            tasks.append(task)

    return dict(
        programs=programs,
        tasks=tasks,
        anonymous_to_named=anonymous_to_named
    )

def rewrite(
    programs: List[str],
    abstractions: List[Abstraction],
    **kwargs
    ) -> List[str]:
    """
    todo description
    """

    args = " ".join([build_arg(k, v) for k, v in kwargs.items()])

    return rewrite_backend(
        programs,
        abstractions,
        args
    )

def compress(
    programs: List[str],
    iterations: int,
    max_arity: int = 2,
    threads: int = 1,
    silent: bool = True,
    **kwargs
    ) -> Dict[str,Any]:
    """
    run compression
    """

    tasks = kwargs.pop("tasks", None)
    anonymous_to_named = kwargs.pop("anonymous_to_named", None)

    kwargs.update(dict(
        iterations=iterations,
        max_arity=max_arity,
        threads=threads,
        silent=silent
    ))

    args = " ".join([build_arg(k, v) for k, v in kwargs.items()])

    try:
        res = compress_backend(
            programs,
            tasks,
            anonymous_to_named,
            args)
    except BaseException as e:
        if e.__class__.__name__ == "PanicException":
            raise StitchException(f"Rust backend panicked with exception: {e}")
        else:
            raise # eg TypeError from pyo3 conversion
    
    res = json.loads(res)
    
    return CompressionResult(res)
    

def build_arg(name: str, val) -> str:
    """
    Builds command line argument version of a Python argument, so for example:
    - build_arg("max_arity",3) -> "--max-arity=3"
    - build_arg("silent",True) -> "--silent"
    - build_arg("silent",False) -> ""
    """
    assert isinstance(val, (str, int, bool)), f"unexpected type for argument `{name}`: {type(val)}"
    res = "--" + name.replace("_", "-")
    if isinstance(val,bool):
        if val:
            return res # eg "--silent"
        else:
            return ""
    return f"{res}={val}"